use std::{error::Error, marker::PhantomData, rc::Rc};

use crate::core::*;

pub trait Atomic: 'static + Sized + Clone + Serializable {
    type ParseError: Error;
    fn deserialize(deserializer: &mut dyn Deserializer) -> Result<Self, Self::ParseError>;
    fn unexpected_tail(tail: &[u8]) -> Self::ParseError;
}

pub struct AtomicFactory<A: Atomic>(PhantomData<A>);

impl<A: Atomic> AtomicFactory<A> {
    fn new() -> Self {
        AtomicFactory(PhantomData)
    }
}

impl<A: Atomic> Clone for AtomicFactory<A> {
    fn clone(&self) -> Self {
        Self::new()
    }
}

impl<'a, Ctx: 'a + Context, A: Atomic> Mentionable<'a, Ctx> for A {
    type Fctr = AtomicFactory<A>;

    fn factory(&self) -> Self::Fctr {
        AtomicFactory::new()
    }

    fn topology(&self) -> Hash {
        Ctx::hash(b"")
    }

    fn points(&self) -> Vec<Point<'a, Ctx, TypelessMentionable<'a, Ctx>>> {
        vec![]
    }
}

impl<'a, Ctx: 'a + Context, A: Atomic> Factory<'a, Ctx> for AtomicFactory<A> {
    type Mtbl = A;

    type ParseError = A::ParseError;

    fn deserialize(
        &self,
        deserializer: &mut dyn Deserializer,
        _resolver: Rc<dyn Resolver<'a, Ctx>>,
    ) -> ParseResult<'a, Ctx, Self> {
        A::deserialize(deserializer)
    }

    fn unexpected_tail(&self, tail: &[u8]) -> Self::ParseError {
        A::unexpected_tail(tail)
    }
}
